import List
import Ratio
import Debug.Trace

newtype Clause a = Clause ([a],[a]) deriving (Eq,Show)

class (Eq f,Show f) => Logic f where 
	ff :: f; tt :: f;
	x :: f; y :: f; z :: f;
	neg :: f -> f
	(/\) :: f -> f -> f
	(\/) :: f -> f -> f
	onlybox :: f -> [f]
	disj :: [f] -> f
	ma :: f -> [f]
	eval :: f -> Clause f -> Bool
	provable :: f -> Bool

class (Logic f,Logic g) => Strip f g where
	strip :: f -> [g]

data L0 = F0 | T0 | Atom0 Int | Neg0 L0 | And0 L0 L0 | Or0 L0 L0 | HM0 Char L1 deriving (Eq,Show)
data L1 = F1 | T1 | Atom1 Int | Neg1 L1 | And1 L1 L1 | Or1 L1 L1 | P1 Rational L0 deriving (Eq,Show)

instance Strip L0 L1 where
	strip (HM0 _ x) = [x]; strip _ = [];

instance Strip L1 L0 where
	strip (P1 _ x) = [x]; strip _ = [];

instance Logic L0 where
	ff = F0; tt = T0;
	x = Atom0 1; y = Atom0 2; z = Atom0 3;
	neg F0 = T0; neg T0 = F0; neg x = Neg0 x
	x \/ y = Or0 x y
	x /\ y = And0 x y
	
	onlybox (HM0 ind x) = [HM0 ind x]; onlybox _ = []

	disj [] = F0; disj [a] = a; disj (a:as) = Or0 a (disj as)

  -- ma :: L0 -> [L0]
	ma (And0 phi psi) = (ma phi) `union` (ma psi)
	ma (Or0 phi psi) = (ma phi) `union` (ma psi)
	ma (Neg0 phi) = ma phi
	ma (T0) = []; ma F0 = []; ma (Atom0 x) = []
	ma x = [x]
	eval (And0 phi psi) s = (eval phi s) && (eval psi s)
	eval (Or0 phi psi) s = (eval phi s) || (eval psi s)
	eval (Neg0 phi) s  = not (eval phi s)
	eval T0 s = True; eval F0 s = False
	eval phi (Clause (pos, neg))
	  | elem phi neg = False
	  | elem phi pos = True
	
	provable phi = all (\c -> any (all provable) (match0 c)) (cnf phi) 


instance Logic L1 where
	ff = F1; tt = T1;
	x = Atom1 1; y = Atom1 2; z = Atom1 3;
	neg F1 = T1; neg T1 = F1; neg x = Neg1 x
	x \/ y = Or1 x y
	x /\ y = And1 x y
	
	onlybox (P1 ind x) = [P1 ind x]; onlybox _ = []

	disj [] = F1; disj [a] = a; disj (a:as) = Or1 a (disj as)
	ma (And1 phi psi) = (ma phi) `union` (ma psi)
	ma (Or1 phi psi) = (ma phi) `union` (ma psi)
	ma (Neg1 phi) = ma phi
	ma F1 = []; ma T1 = [];
	ma x = [x]
	eval (And1 phi psi) s = (eval phi s) && (eval psi s)
	eval (Or1 phi psi) s = (eval phi s) || (eval psi s)
	eval (Neg1 phi) s  = not (eval phi s)
	eval T1 s = True; eval F1 s = False
	eval phi (Clause (pos, neg))
	  | elem phi neg = False
	  | elem phi pos = True
	
	provable phi = all (\c -> any (all provable) (match1 c)) (cnf phi) 


match0 :: Clause L0 -> [[L1]]
match0 (Clause (pl,nl)) =
  let isag a (HM0 ind x) = (ind == a) -- is it agent a?
      -- isag a _ = False
      -- debug: there seem to be cases when we are fed a ``F0'' i
      -- in c clause!
      ag (HM0 ind x) = ind; -- find the agent
      ag _ = 'a'
      ags = nub (map ag (pl ++ nl)); -- the set of agents we have
      aglits = map (\x -> (filter (isag x) pl, filter (isag x) nl)) ags;
      -- this is a list of pairs (pl, nl) for each agent
      strlits  = map (\(x, y) -> (map neg (striplst x), striplst y)) aglits;
      -- the same with boxes stripped
      comb (nls, pls) = map disjlst (map (\x -> x:nls) pls);
      -- turn a combination pls/nls into a list of formulas
      fllist = map comb strlits;
      -- a list of lists consisting of formulas
  in concat fllist

match1 :: Clause L1 -> [[L0]]
match1 (Clause (pls,nls)) =
		let	(bpls,bnls) = (onlyboxes pls, onlyboxes nls)
			allcls = [Clause(ps,ns) | 
				ps <- (pwrset bpls), ns <- (pwrset bnls)] \\ [Clause([],[])]
			xrats xs = map (\(P1 k _) -> k) xs
			bnd (Clause(ps,ns)) = pmlbnd ((xrats ps),(xrats ns))
			tups cl@(Clause(ps,ns)) = 
				[(pts,nts,k)|	pts <- (tuprange (bnd cl) (length ps)), 
								nts <- (tuprange (bnd cl) (length ns)),
								k <- [-(bnd cl)..(bnd cl)]]
			sctups cl = filter (pmlsc1 cl) (tups cl)
			clmatch cl = map (pmlcnf1 cl) (sctups cl) 
		in	concatMap clmatch allcls

pmlcnf1 :: Clause L1 -> ([Int],[Int], Int) -> [L0]
pmlcnf1 (Clause(pls,nls)) (prs,nrs,k) =
	let	(pinds,ninds) = (toinds prs,toinds nrs)
		(spls,snls) = (striplst pls,striplst nls)
		allinds = [(ps,ns) | ps <- (pwrset pinds), ns <- (pwrset ninds)]
		cnfinds = filter (\(ps,ns) -> (sum$img ps prs) - (sum$img ns nrs) < k) allinds
		getJ (ps,ns) = (img ps spls) ++ (img ns snls)
		getnJ (ps,ns) = (img (pinds \\ ps) spls) ++ (img (ninds \\ ns) snls)
	in	map (\rs -> ndisj(getJ rs) \/ disj(getnJ rs)) cnfinds

pmlsc1 :: Clause L1 -> ([Int],[Int], Int) -> Bool
pmlsc1 (Clause(pls,nls)) (pints,nints,k) =
	let	(rpints,rnints) = (map fromIntegral pints,map fromIntegral nints)
		psum = sum $ zipbin (*) rpints (map (\(P1 x _)->fromRational(x)) pls)
		nsum = sum $ zipbin (*) rnints (map (\(P1 x _)-> fromRational(-x)) nls)
	in	if null(pints) 	then (psum + nsum < fromIntegral(k))
						else (psum + nsum <= fromIntegral(k))

pmlbnd :: ([Rational],[Rational]) -> Int
pmlbnd _ = 1
pmlbnd (rps,rns) = 
	let	n = (length rps) + (length rns)
		toints rs = concatMap (\r -> [numerator r,denominator r]) rs
		allints = (toints rps) ++ (toints rns)
		logint n = ceiling(logBase 2 (1 + n))
		logsum = sum $ map (\n -> logint (fromIntegral(n))) allints
	in	20*n*n*(1+n) + 10*n*n*logsum


-- generic code (no need for duplication)
------------------------------------------

(-->) :: (Logic f) => f -> f -> f
x --> y = (neg x) \/ y

(<-->) :: (Logic f) => f -> f -> f
x <--> y = (x --> y) /\ (y --> x)

-- disjunction of negation of elements
ndisj :: (Logic f) => [f] -> f
ndisj xs = disj (map neg xs)


-- create singleton lists containing disjunctions
disjlst :: (Logic f) => [f] -> [f]
disjlst ls = [disj ls]

-- negate a list
neglst :: (Logic f) => [f] -> [f]
neglst xs = map neg xs 

-- generic: purely generic

-- avoids []->[[]], otherwise F matches F
listify :: [a] -> [[a]]
listify [] = []; listify x = [x]

-- length ordering on lists
cmpsize :: [a] -> [a] -> Ordering
cmpsize s t = if(length(s) < length(t)) then LT else GT

-- zip two lists together using a binary operator
zipbin :: (a -> a -> a) -> [a] -> [a] -> [a]
zipbin _ [] _ = [];	zipbin _ _ [] = [];
zipbin f (x:xs) (y:ys) = (f x y):(zipbin f xs ys)

-- given indices and a list find image
img :: [Int] -> [a] -> [a]
img inds ls = map (ls!!) inds 

-- map a list to indices from 0 -> length -1
toinds :: [a] -> [Int]
toinds [] = []; toinds xs = [0..(length xs)-1]


-- construct all integer n-tuples with elements from 1,..,r
tuprange :: Int -> Int -> [[Int]]
tuprange _ 0 = [[]]
tuprange r n = 
	let rec xs ys = map (\z -> z:ys) xs
	in	concatMap (rec [1..r]) (tuprange r (n-1))

	
-- generic: set operations

pwdisj :: (Eq a) => [[a]] -> Bool
pwdisj [] = True
pwdisj (x:xs) = if (all (\y -> null(x `intersect` y)) xs) then (pwdisj xs) else False
	
pwrset :: [a] -> [[a]]
pwrset [] = [[]]; pwrset (x:xs) = (map ([x]++) (pwrset xs)) ++ (pwrset xs) 

-- remove subsets
rmsub :: (Eq a) => [[a]] -> [[a]]
rmsub [] = []
rmsub (x:xs) | (any (\y -> (x `intersect` y == x)) xs) = rmsub(xs) | otherwise = x:rmsub(xs) 


-- generic: strippers and box collecting

-- strip a list
striplst :: (Strip f g) => [f] -> [g]
striplst ls = concatMap strip ls

onlyboxes :: (Logic f) => [f] -> [f]
onlyboxes ls = concatMap onlybox ls

cml :: (a -> [a]) -> [a] -> [a]
cml f ls = concatMap f ls

-- generic: main prover

-- all possible clauses
clauses :: (Logic f) => [f] -> [Clause f]
clauses [] = [Clause ([],[])]
clauses (a:as) = 
	let	padd c (Clause (p, n)) = Clause (c:p, n)
		nadd c (Clause (p, n)) = Clause (p, c:n)
	in	(map (padd a) (clauses as)) ++ (map (nadd a) (clauses as))
	
-- dnf
allsat :: (Logic f) => f -> [Clause f]
allsat phi = filter (\x -> (eval phi x)) (clauses (ma phi))

-- cnf
cnf :: (Logic f) => f -> [Clause f]
cnf phi = map (\(Clause (x, y)) -> (Clause (y, x))) (allsat (neg phi))

-- sat
sat :: (Logic f) => f -> Bool
sat phi = not $ provable $ neg phi

-- examples from the paper written using some syntactic sugar:
dia :: Char -> L1 -> L0; dia c phi = neg (HM0 c (neg phi))
l :: Rational -> L0 -> L1; l r phi = P1 r phi
phi1 :: L0; phi1 = dia 'i'  ((l (7 % 10) (dia 'o' tt)) /\  (l (3 % 10) (dia 'e'  tt)))
psi :: L0; psi = dia 'i' (l (9%10) (dia 'o' tt))

main = putStr (show (provable phi1))
